@@ -26,6 +26,7 @@
 #include <string.h>
 #include <errno.h>
 #include <dirent.h>
+#include <sys/types.h>
 #include <sys/stat.h>
 
 #include "ioconf.h"
@@ -41,6 +42,7 @@
 
 static unsigned int ioc_parsed = 0;
 static struct ioc_entry *ioconf[MAX_BLKDEV + 1];
+static unsigned int ioc_refnr[MAX_BLKDEV + 1];
 
 /*
  ***************************************************************************
@@ -140,18 +142,29 @@ int ioc_init(void)
 {
 	FILE *fp;
 	unsigned int i, major, indirect, count = 0;
-	char buf[IOC_LINESIZ + 1];
-	char cfmt[IOC_FMTLEN + 1];
-	char dfmt[IOC_FMTLEN + 1];
-	char pfmt[IOC_FMTLEN + 1];
-	char desc[IOC_DESCLEN + 1];
+	char buf[IOC_LINESIZ];
+	char cfmt[IOC_FMTLEN];
+	char dfmt[IOC_FMTLEN];
+	char pfmt[IOC_FMTLEN];
+	char desc[IOC_DESCLEN];
 	struct ioc_entry  *iocp = NULL;
 	struct blk_config *blkp = NULL;
+	char ioconf_name[64];
 
-	if ((fp = fopen(IOCONF, "r")) == NULL)
-		return 0;
+	if ((fp = fopen(IOCONF, "r")) == NULL) {
+		if ((fp = fopen(LOCAL_IOCONF, "r")) == NULL)
+			return 0;
+		strncpy(ioconf_name, LOCAL_IOCONF, 64);
+	}
+	else {
+		strncpy(ioconf_name, IOCONF, 64);
+	}
+	ioconf_name[63] = '\0';
 
-	while (fgets(buf, IOC_LINESIZ, fp)) {
+	/* Init ioc_refnr array */
+	memset(ioc_refnr, 0, sizeof(ioc_refnr));
+
+	while (fgets(buf, IOC_LINESIZ - 1, fp)) {
 
 		if ((*buf == '#') || (*buf == '\n'))
 			continue;
@@ -178,16 +191,21 @@ int ioc_init(void)
 				/* conventional usage for unsupported device */
 				continue;
 			}
-			if (indirect >= MAX_BLKDEV) {
+			if (indirect > MAX_BLKDEV) {
 				fprintf(stderr, "%s: Indirect major #%u out of range\n",
-					IOCONF, indirect);
+					ioconf_name, indirect);
+				continue;
+			}
+			if (major > MAX_BLKDEV) {
+				fprintf(stderr, "%s: Major #%u out of range\n",
+					ioconf_name, major);
 				continue;
 			}
 			if (ioconf[indirect] == NULL) {
 				fprintf(stderr,
 					"%s: Indirect record '%u:%u:%u:...'"
 					" references not yet seen major %u\n",
-					IOCONF, major, indirect, iocp->ctrlno, major);
+					ioconf_name, major, indirect, iocp->ctrlno, major);
 				continue;
 			}
 			/*
@@ -201,10 +219,12 @@ int ioc_init(void)
 				iocp->desc = ioconf[indirect]->blkp->desc;
 			}
 			else {
-				IOC_ALLOC(iocp->desc, char, IOC_DESCLEN + 1);
-				strncpy(iocp->desc, desc, IOC_DESCLEN);
+				IOC_ALLOC(iocp->desc, char, IOC_DESCLEN);
+				memcpy(iocp->desc, desc, IOC_DESCLEN);
 			}
+			ioc_refnr[indirect]++;
 			ioconf[major] = iocp;
+			iocp->basemajor = indirect;
 			iocp->blkp = ioconf[indirect]->blkp;
 			iocp->live = 0;
 			iocp = NULL;
@@ -214,7 +234,7 @@ int ioc_init(void)
 
 		/* maybe it's a full record? */
 
-		i = sscanf(buf, "%u:%[^:]:%[^:]:%d:%[^:]:%u:%[^:]:%u:%s",
+		i = sscanf(buf, "%u:%[^:]:%[^:]:%u:%[^:]:%u:%[^:]:%u:%s",
 			   &major, blkp->name,
 			   cfmt, &iocp->ctrlno,
 			   dfmt, &blkp->dcount,
@@ -222,8 +242,8 @@ int ioc_init(void)
 			   desc);
 
 		if (i != 9) {
-			fprintf(stderr, "%s: Malformed %d field record: %s\n",
-				IOCONF, i, buf);
+			fprintf(stderr, "%s: Malformed %u field record: %s\n",
+				ioconf_name, i, buf);
 			continue;
 		}
 
@@ -241,13 +261,13 @@ int ioc_init(void)
 
 			/*
 			 * device has an aliased minor
-			 * for now we only support on exception per major
+			 * for now we only support one exception per major
 			 * (catering to initrd: (1,250))
 			 */
 			if (ioconf[major] == NULL) {
 				fprintf(stderr, "%s: type 'x' record for"
 					" major #%u must follow the base record - ignored\n",
-					IOCONF, major);
+					ioconf_name, major);
 				continue;
 			}
 			xblkp = ioconf[major]->blkp;
@@ -260,7 +280,7 @@ int ioc_init(void)
 				 */
 				fprintf(stderr, "%s: duplicate 'x' record for"
 					" major #%u - ignored\ninput line: %s\n",
-					IOCONF, major, buf);
+					ioconf_name, major, buf);
 				continue;
 			}
 			/*
@@ -268,7 +288,8 @@ int ioc_init(void)
 			 * exception info
 			 */
 			xblkp->ext_minor = iocp->ctrlno;
-			strcpy(xblkp->ext_name, blkp->name);
+			strncpy(xblkp->ext_name, blkp->name, IOC_NAMELEN);
+			xblkp->ext_name[IOC_NAMELEN - 1] = '\0';
 			xblkp->ext = 1;
 			continue;
 		}
@@ -280,7 +301,8 @@ int ioc_init(void)
 
 		/* basename of device + provided string + controller # */
 		if (*cfmt == '*') {
-			strcpy(blkp->cfmt, blkp->name);
+			strncpy(blkp->cfmt, blkp->name, sizeof(blkp->cfmt) - 1);
+			blkp->cfmt[sizeof(blkp->cfmt) - 1] = '\0';
 		}
 		else {
 			sprintf(blkp->cfmt, "%s%s%%d", blkp->name, cfmt);
@@ -296,7 +318,8 @@ int ioc_init(void)
 			break;
 
 		case '%':
-			strcpy(blkp->dfmt, dfmt + 1);
+			strncpy(blkp->dfmt, dfmt + 1, IOC_FMTLEN - 1);
+			/* fallthrough to next case */
 		case 'd':
 			blkp->cconv = ioc_ito10;
 			strcat(blkp->dfmt, "%s");
@@ -313,8 +336,9 @@ int ioc_init(void)
 		iocp->live = 1;
 		iocp->blkp = blkp;
 		iocp->desc = NULL;
+		iocp->basemajor = major;
 		ioconf[major] = iocp;
-		strncpy(blkp->desc, desc, IOC_DESCLEN);
+		memcpy(blkp->desc, desc, IOC_DESCLEN);
 		blkp = NULL; iocp = NULL;
 		++count;
 	}
@@ -323,17 +347,27 @@ int ioc_init(void)
 	/*
 	 * These will become leaks if we ever 'continue'
 	 * after IOC_ALLOC( blkp->desc ... ).
-	 * Right Now, we don't.
+	 * Right now, we don't.
 	 */
-	if (blkp != NULL)
-		free(blkp);
-	if (iocp != NULL)
-		free(iocp);
+	free(blkp);
+	free(iocp);
 
 	/* Indicate that ioconf file has been parsed */
 	ioc_parsed = 1;
 
 	return (count);
+
+free_and_return:
+	/* Free pointers and return */
+	fclose(fp);
+	if (blkp) {
+		free(blkp);
+	}
+	if (iocp) {
+		free(iocp);
+	}
+
+	return 0;
 }
 
 /*
@@ -353,11 +387,11 @@ int ioc_init(void)
 
 char *ioc_name(unsigned int major, unsigned int minor)
 {
-	static char name[IOC_DEVLEN + 1];
+	static char name[IOC_DEVLEN];
 	struct ioc_entry *p;
 	int base, offset;
 
-	if ((MAX_BLKDEV <= major) || (IOC_MAXMINOR <= minor)) {
+	if ((major > MAX_BLKDEV) || (minor > IOC_MAXMINOR)) {
 		return (NULL);
 	}
 
@@ -366,9 +400,8 @@ char *ioc_name(unsigned int major, unsigned int minor)
 
 	p = ioconf[major];
 
-
 	/* Invalid major or minor numbers? */
-	if ((p == NULL) || (minor >= (p->blkp->dcount * p->blkp->pcount))) {
+	if ((p == NULL) || ((minor & 0xff) >= (p->blkp->dcount * p->blkp->pcount))) {
 		/*
 		 * That minor test is only there for IDE-style devices
 		 * that have no minors over 128.
@@ -379,7 +412,8 @@ char *ioc_name(unsigned int major, unsigned int minor)
 
 	/* Is this an extension record? */
 	if (p->blkp->ext && (p->blkp->ext_minor == minor)) {
-		strcpy(name, p->blkp->ext_name);
+		strncpy(name, p->blkp->ext_name, IOC_DEVLEN);
+		name[IOC_DEVLEN - 1] = '\0';
 		return (name);
 	}
 
@@ -392,9 +426,14 @@ char *ioc_name(unsigned int major, unsigned int minor)
 	 * (we parse once but may generate lots of names)
 	 */
 	base = p->ctrlno * p->blkp->dcount;
-	offset = minor / p->blkp->pcount;
-	if (!p->blkp->ctrl_explicit)
+	if (minor >= 256) {
+		base += p->blkp->dcount * (ioc_refnr[p->basemajor] + 1) * (minor >> 8);
+	}
+
+	offset = (minor & 0xff) / p->blkp->pcount;
+	if (!p->blkp->ctrl_explicit) {
 		offset += base;
+	}
 
 	/*
 	 * These sprintfs can't be coalesced because the first might
@@ -431,16 +470,12 @@ int ioc_iswhole(unsigned int major, unsigned int minor)
 	if (!ioc_parsed && !ioc_init())
 		return 0;
 
-	if (major >= MAX_BLKDEV)
-		/*
-		 * Later: Handle Linux long major numbers here.
-		 * Now: This is an error.
-		 */
+	if (major > MAX_BLKDEV)
 		return 0;
 
 	if (ioconf[major] == NULL)
 		/* Device not registered */
-		return 0 ;
+		return 0;
 
 	return (IS_WHOLE(major, minor));
 }
@@ -464,6 +499,7 @@ char *transform_devmapname(unsigned int major, unsigned int minor)
 	struct dirent *dp;
 	char filen[MAX_FILE_LEN];
 	char *dm_name = NULL;
+	static char name[MAX_NAME_LEN];
 	struct stat aux;
 	unsigned int dm_major, dm_minor;
 
@@ -478,14 +514,16 @@ char *transform_devmapname(unsigned int major, unsigned int minor)
 		snprintf(filen, MAX_FILE_LEN, "%s/%s", DEVMAP_DIR, dp->d_name);
 		filen[MAX_FILE_LEN - 1] = '\0';
 
-		if (stat(filen, &aux) == 0) {
+		if (__stat(filen, &aux) == 0) {
 			/* Get its minor and major numbers */
 
-			dm_major = ((aux.st_rdev >> 8) & 0xff);
-			dm_minor = (aux.st_rdev & 0xff);
-	
+			dm_major = major(aux.st_rdev);
+			dm_minor = minor(aux.st_rdev);
+
 			if ((dm_minor == minor) && (dm_major == major)) {
-				dm_name = dp->d_name;
+				strncpy(name, dp->d_name, MAX_NAME_LEN);
+				name[MAX_NAME_LEN - 1] = '\0';
+				dm_name = name;
 				break;
 			}
 		}
