@@ -1,6 +1,6 @@
 /*
  * sar and sadf common routines.
- * (C) 1999-2018 by Sebastien GODARD (sysstat <at> orange.fr)
+ * (C) 1999-2019 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -28,6 +28,7 @@
 #include <unistd.h>	/* For STDOUT_FILENO, among others */
 #include <dirent.h>
 #include <fcntl.h>
+#include <limits.h>
 #include <libgen.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -52,6 +53,7 @@ extern unsigned int dm_major;
 unsigned int hdr_types_nr[] = {FILE_HEADER_ULL_NR, FILE_HEADER_UL_NR, FILE_HEADER_U_NR};
 unsigned int act_types_nr[] = {FILE_ACTIVITY_ULL_NR, FILE_ACTIVITY_UL_NR, FILE_ACTIVITY_U_NR};
 unsigned int rec_types_nr[] = {RECORD_HEADER_ULL_NR, RECORD_HEADER_UL_NR, RECORD_HEADER_U_NR};
+unsigned int extra_desc_types_nr[] = {EXTRA_DESC_ULL_NR, EXTRA_DESC_UL_NR, EXTRA_DESC_U_NR};
 unsigned int nr_types_nr[]  = {0, 0, 1};
 
 /*
@@ -143,6 +145,7 @@ void guess_sa_name(char *sa_dir, struct tm *rectime, int *sa_name)
 	char filename[MAX_FILE_LEN];
 	struct stat sb;
 	time_t sa_mtime;
+	long nsec;
 
 	/* Use saDD by default */
 	*sa_name = 0;
@@ -159,6 +162,7 @@ void guess_sa_name(char *sa_dir, struct tm *rectime, int *sa_name)
 		/* Cannot find or access saYYYYMMDD, so use saDD */
 		return;
 	sa_mtime = sb.st_mtime;
+	nsec = sb.st_mtim.tv_nsec;
 
 	/* Look for saDD */
 	snprintf(filename, MAX_FILE_LEN,
@@ -172,7 +176,8 @@ void guess_sa_name(char *sa_dir, struct tm *rectime, int *sa_name)
 		return;
 	}
 
-	if (sa_mtime > sb.st_mtime) {
+	if ((sa_mtime > sb.st_mtime) ||
+	    ((sa_mtime == sb.st_mtime) && (nsec > sb.st_mtim.tv_nsec))) {
 		/* saYYYYMMDD is more recent than saDD, so use it */
 		*sa_name = 1;
 	}
@@ -240,6 +245,10 @@ int set_default_file(char *datafile, int d_off, int sa_name)
 	datafile[MAX_FILE_LEN - 1] = '\0';
 	default_file_used = TRUE;
 
+#ifdef DEBUG
+	fprintf(stderr, "%s: Datafile: %s\n", __FUNCTION__, datafile);
+#endif
+
 	return ((err < 0) || (err >= MAX_FILE_LEN));
 }
 
@@ -535,11 +544,11 @@ void reallocate_all_buffers(struct activity *a, __nr_t nr_min)
   */
 char *get_devname_from_sysfs(unsigned int major, unsigned int minor)
 {
-	static char link[32], target[PATH_MAX];
+	static char link[256], target[PATH_MAX];
 	char *devname;
 	ssize_t r;
 
-	snprintf(link, 32, "%s/%u:%u", SYSFS_DEV_BLOCK, major, minor);
+	snprintf(link, 256, "%s/%u:%u", SYSFS_DEV_BLOCK, major, minor);
 
 	/* Get full path to device knowing its major and minor numbers */
 	r = readlink(link, target, PATH_MAX);
@@ -699,22 +708,34 @@ int decode_timestamp(char timestamp[], struct tstamp *tse)
  * IN:
  * @rectime	Date and time for current sample.
  * @tse		Timestamp used as reference.
+ * @cross_day	TRUE if a new day has been started.
  *
  * RETURNS:
  * A positive value if @rectime is greater than @tse,
  * a negative one otherwise.
  ***************************************************************************
  */
-int datecmp(struct tm *rectime, struct tstamp *tse)
+int datecmp(struct tm *rectime, struct tstamp *tse, int cross_day)
 {
-	if (rectime->tm_hour == tse->tm_hour) {
+	int tm_hour = rectime->tm_hour;
+
+	if (cross_day) {
+		/*
+		 * This is necessary if we want to properly handle something like:
+		 * sar -s time_start -e time_end with
+		 * time_start(day D) > time_end(day D+1)
+		 */
+		tm_hour += 24;
+	}
+
+	if (tm_hour == tse->tm_hour) {
 		if (rectime->tm_min == tse->tm_min)
 			return (rectime->tm_sec - tse->tm_sec);
 		else
 			return (rectime->tm_min - tse->tm_min);
 	}
 	else
-		return (rectime->tm_hour - tse->tm_hour);
+		return (tm_hour - tse->tm_hour);
 }
 
 /*
@@ -801,11 +822,9 @@ void get_itv_value(struct record_header *record_hdr_curr,
  * 		not the time, should be used by the caller.
  ***************************************************************************
  */
-void get_file_timestamp_struct(unsigned int flags, struct tm *rectime,
+void get_file_timestamp_struct(uint64_t flags, struct tm *rectime,
 			       struct file_header *file_hdr)
 {
-	struct tm *loc_t;
-
 	if (PRINT_TRUE_TIME(flags)) {
 		/* Get local time. This is just to fill fields with a default value. */
 		get_time(rectime, 0);
@@ -821,9 +840,7 @@ void get_file_timestamp_struct(unsigned int flags, struct tm *rectime,
 		mktime(rectime);
 	}
 	else {
-		if ((loc_t = localtime((const time_t *) &file_hdr->sa_ust_time)) != NULL) {
-			*rectime = *loc_t;
-		}
+		localtime_r((const time_t *) &file_hdr->sa_ust_time, rectime);
 	}
 }
 
@@ -839,7 +856,7 @@ void get_file_timestamp_struct(unsigned int flags, struct tm *rectime,
  * @rectime	Date and time from file header.
  ***************************************************************************
  */
-void print_report_hdr(unsigned int flags, struct tm *rectime,
+void print_report_hdr(uint64_t flags, struct tm *rectime,
 		      struct file_header *file_hdr)
 {
 
@@ -1094,11 +1111,13 @@ int check_disk_reg(struct activity *a, int curr, int ref, int pos)
 			 * If only one or two have decreased then the likelyhood
 			 * is that the counter has simply wrapped.
 			 * Don't take into account a counter if its previous value was 0
-			 * (this may be a read-only device).
+			 * (this may be a read-only device, or a kernel that doesn't
+			 * support discard stats yet...)
 			 */
 			if ((sdc->nr_ios < sdp->nr_ios) &&
 			    (!sdp->rd_sect || (sdc->rd_sect < sdp->rd_sect)) &&
-			    (!sdp->wr_sect || (sdc->wr_sect < sdp->wr_sect)))
+			    (!sdp->wr_sect || (sdc->wr_sect < sdp->wr_sect)) &&
+			    (!sdp->dc_sect || (sdc->dc_sect < sdp->dc_sect)))
 				/* Same device registered again */
 				return -2;
 
@@ -1283,27 +1302,35 @@ void swap_struct(unsigned int types_nr[], void *ps, int is64bit)
  *		size of the structure *read from file*.
  * @g_size	Size of the structure expected by current sysstat version.
  * @b_size	Size of the buffer pointed by @ps.
+ *
+ * RETURNS:
+ * -1 if an error has been encountered, or 0 otherwise.
  ***************************************************************************
  */
-void remap_struct(unsigned int gtypes_nr[], unsigned int ftypes_nr[],
-		  void *ps, unsigned int f_size, unsigned int g_size, size_t b_size)
+int remap_struct(unsigned int gtypes_nr[], unsigned int ftypes_nr[],
+		 void *ps, unsigned int f_size, unsigned int g_size, size_t b_size)
 {
 	int d;
 	size_t n;
 
 	/* Sanity check */
 	if (MAP_SIZE(ftypes_nr) > f_size)
-		return;
+		return -1;
 
 	/* Remap [unsigned] long fields */
 	d = gtypes_nr[0] - ftypes_nr[0];
 	if (d) {
+		if (ftypes_nr[0] * ULL_ALIGNMENT_WIDTH < ftypes_nr[0])
+			/* Overflow */
+			return -1;
+
 		n = MINIMUM(f_size - ftypes_nr[0] * ULL_ALIGNMENT_WIDTH,
 			    g_size - gtypes_nr[0] * ULL_ALIGNMENT_WIDTH);
 		if ((ftypes_nr[0] * ULL_ALIGNMENT_WIDTH >= b_size) ||
 		    (gtypes_nr[0] * ULL_ALIGNMENT_WIDTH + n > b_size) ||
 		    (ftypes_nr[0] * ULL_ALIGNMENT_WIDTH + n > b_size))
-			return;
+			return -1;
+
 		memmove(((char *) ps) + gtypes_nr[0] * ULL_ALIGNMENT_WIDTH,
 			((char *) ps) + ftypes_nr[0] * ULL_ALIGNMENT_WIDTH, n);
 		if (d > 0) {
@@ -1314,6 +1341,11 @@ void remap_struct(unsigned int gtypes_nr[], unsigned int ftypes_nr[],
 	/* Remap [unsigned] int fields */
 	d = gtypes_nr[1] - ftypes_nr[1];
 	if (d) {
+		if (gtypes_nr[0] * ULL_ALIGNMENT_WIDTH +
+		    ftypes_nr[1] * UL_ALIGNMENT_WIDTH < ftypes_nr[1])
+			/* Overflow */
+			return -1;
+
 		n = MINIMUM(f_size - ftypes_nr[0] * ULL_ALIGNMENT_WIDTH
 				   - ftypes_nr[1] * UL_ALIGNMENT_WIDTH,
 			    g_size - gtypes_nr[0] * ULL_ALIGNMENT_WIDTH
@@ -1324,7 +1356,8 @@ void remap_struct(unsigned int gtypes_nr[], unsigned int ftypes_nr[],
 		     gtypes_nr[1] * UL_ALIGNMENT_WIDTH + n > b_size) ||
 		    (gtypes_nr[0] * ULL_ALIGNMENT_WIDTH +
 		     ftypes_nr[1] * UL_ALIGNMENT_WIDTH + n > b_size))
-			return;
+			return -1;
+
 		memmove(((char *) ps) + gtypes_nr[0] * ULL_ALIGNMENT_WIDTH
 				      + gtypes_nr[1] * UL_ALIGNMENT_WIDTH,
 			((char *) ps) + gtypes_nr[0] * ULL_ALIGNMENT_WIDTH
@@ -1338,6 +1371,12 @@ void remap_struct(unsigned int gtypes_nr[], unsigned int ftypes_nr[],
 	/* Remap possible fields (like strings of chars) following int fields */
 	d = gtypes_nr[2] - ftypes_nr[2];
 	if (d) {
+		if (gtypes_nr[0] * ULL_ALIGNMENT_WIDTH +
+		    gtypes_nr[1] * UL_ALIGNMENT_WIDTH +
+		    ftypes_nr[2] * U_ALIGNMENT_WIDTH < ftypes_nr[2])
+			/* Overflow */
+			return -1;
+
 		n = MINIMUM(f_size - ftypes_nr[0] * ULL_ALIGNMENT_WIDTH
 				   - ftypes_nr[1] * UL_ALIGNMENT_WIDTH
 				   - ftypes_nr[2] * U_ALIGNMENT_WIDTH,
@@ -1353,7 +1392,8 @@ void remap_struct(unsigned int gtypes_nr[], unsigned int ftypes_nr[],
 		    (gtypes_nr[0] * ULL_ALIGNMENT_WIDTH +
 		     gtypes_nr[1] * UL_ALIGNMENT_WIDTH +
 		     ftypes_nr[2] * U_ALIGNMENT_WIDTH + n > b_size))
-			return;
+			return -1;
+
 		memmove(((char *) ps) + gtypes_nr[0] * ULL_ALIGNMENT_WIDTH
 				      + gtypes_nr[1] * UL_ALIGNMENT_WIDTH
 				      + gtypes_nr[2] * U_ALIGNMENT_WIDTH,
@@ -1367,6 +1407,7 @@ void remap_struct(unsigned int gtypes_nr[], unsigned int ftypes_nr[],
 			       0, d * U_ALIGNMENT_WIDTH);
 		}
 	}
+	return 0;
 }
 
 /*
@@ -1379,12 +1420,16 @@ void remap_struct(unsigned int gtypes_nr[], unsigned int ftypes_nr[],
  * @size	Number of bytes to read.
  * @mode	If set to HARD_SIZE, indicate that an EOF should be considered
  * 		as an error.
+ * @oneof	Set to UEOF_CONT if an unexpected end of file should not make
+ *		sadf stop. Default behavior is to stop on unexpected EOF.
  *
  * RETURNS:
- * 1 if EOF has been reached, 0 otherwise.
+ * 1 if EOF has been reached,
+ * 2 if an unexpected EOF has been reached (and sadf was told to continue),
+ * 0 otherwise.
  ***************************************************************************
  */
-int sa_fread(int ifd, void *buffer, size_t size, int mode)
+int sa_fread(int ifd, void *buffer, size_t size, int mode, int oneof)
 {
 	ssize_t n;
 
@@ -1400,6 +1445,8 @@ int sa_fread(int ifd, void *buffer, size_t size, int mode)
 
 	if (n < size) {
 		fprintf(stderr, _("End of system activity file unexpected\n"));
+		if (oneof == UEOF_CONT)
+			return 2;
 		close(ifd);
 		exit(2);
 	}
@@ -1407,6 +1454,66 @@ int sa_fread(int ifd, void *buffer, size_t size, int mode)
 	return 0;
 }
 
+/*
+ ***************************************************************************
+ * Skip unknown extra structures present in file.
+ *
+ * IN:
+ * @ifd		System activity data file descriptor.
+ * @endian_mismatch
+ *		TRUE if file's data don't match current machine's endianness.
+ * @arch_64	TRUE if file's data come from a 64 bit machine.
+ *
+ * RETURNS:
+ * -1 on error, 0 otherwise.
+ ***************************************************************************
+ */
+int skip_extra_struct(int ifd, int endian_mismatch, int arch_64)
+{
+	int i;
+	struct extra_desc xtra_d;
+
+	do {
+		/* Read extra structure description */
+		sa_fread(ifd, &xtra_d, EXTRA_DESC_SIZE, HARD_SIZE, UEOF_STOP);
+
+		/*
+		 * We don't need to remap as the extra_desc structure won't change,
+		 * but we may need to normalize endianness anyway.
+		 */
+		if (endian_mismatch) {
+			swap_struct(extra_desc_types_nr, &xtra_d, arch_64);
+		}
+
+		/* Check values consistency */
+		if (MAP_SIZE(xtra_d.extra_types_nr) > xtra_d.extra_size) {
+#ifdef DEBUG
+			fprintf(stderr, "%s: extra_size=%u types=%d,%d,%d\n",
+				__FUNCTION__, xtra_d.extra_size,
+				xtra_d.extra_types_nr[0], xtra_d.extra_types_nr[1], xtra_d.extra_types_nr[2]);
+#endif
+			return -1;
+		}
+
+		if ((xtra_d.extra_nr > MAX_EXTRA_NR) || (xtra_d.extra_size > MAX_EXTRA_SIZE)) {
+#ifdef DEBUG
+			fprintf(stderr, "%s: extra_size=%u extra_nr=%u\n",
+				__FUNCTION__, xtra_d.extra_size, xtra_d.extra_size);
+#endif
+			return -1;
+		}
+
+		/* Ignore current unknown extra structures */
+		for (i = 0; i < xtra_d.extra_nr; i++) {
+			if (lseek(ifd, xtra_d.extra_size, SEEK_CUR) < xtra_d.extra_size)
+				return -1;
+		}
+	}
+	while (xtra_d.extra_next);
+
+	return 0;
+}
+
 /*
  ***************************************************************************
  * Read the record header of current sample and process it.
@@ -1421,32 +1528,60 @@ int sa_fread(int ifd, void *buffer, size_t size, int mode)
  * @endian_mismatch
  *		TRUE if data read from file don't match current machine's
  *		endianness.
+ * @oneof	Set to EOF_CONT if an unexpected end of file should not make
+ *		sadf stop. Default behavior is to stop on unexpected EOF.
  * @b_size	@buffer size.
+ * @flags	Flags for common options and system state.
  *
  * OUT:
  * @record_hdr	Record header for current sample.
  *
  * RETURNS:
- * 1 if EOF has been reached, 0 otherwise.
+ * 1 if EOF has been reached,
+ * 2 if an error has been encountered (e.g. unexpected EOF),
+ * 0 otherwise.
  ***************************************************************************
  */
 int read_record_hdr(int ifd, void *buffer, struct record_header *record_hdr,
 		    struct file_header *file_hdr, int arch_64, int endian_mismatch,
-		    size_t b_size)
+		    int oneof, size_t b_size, uint64_t flags)
 {
-	if (sa_fread(ifd, buffer, (size_t) file_hdr->rec_size, SOFT_SIZE))
-		/* End of sa data file */
-		return 1;
+	int rc;
 
-	/* Remap record header structure to that expected by current version */
-	remap_struct(rec_types_nr, file_hdr->rec_types_nr, buffer,
-		     file_hdr->rec_size, RECORD_HEADER_SIZE, b_size);
-	memcpy(record_hdr, buffer, RECORD_HEADER_SIZE);
+	do {
+		if ((rc = sa_fread(ifd, buffer, (size_t) file_hdr->rec_size, SOFT_SIZE, oneof)) != 0)
+			/* End of sa data file */
+			return rc;
 
-	/* Normalize endianness */
-	if (endian_mismatch) {
-		swap_struct(rec_types_nr, record_hdr, arch_64);
+		/* Remap record header structure to that expected by current version */
+		if (remap_struct(rec_types_nr, file_hdr->rec_types_nr, buffer,
+				 file_hdr->rec_size, RECORD_HEADER_SIZE, b_size) < 0)
+			return 2;
+		memcpy(record_hdr, buffer, RECORD_HEADER_SIZE);
+
+		/* Normalize endianness */
+		if (endian_mismatch) {
+			swap_struct(rec_types_nr, record_hdr, arch_64);
+		}
+
+		/* Raw output in debug mode */
+		if (DISPLAY_DEBUG_MODE(flags)) {
+			printf("# uptime_cs; %llu; ust_time; %llu; extra_next; %u; record_type; %d; HH:MM:SS; %02d:%02d:%02d\n",
+			       record_hdr->uptime_cs, record_hdr->ust_time,
+			       record_hdr->extra_next, record_hdr->record_type,
+			       record_hdr->hour, record_hdr->minute, record_hdr->second);
+		}
+
+		/*
+		 * Skip unknown extra structures if present.
+		 * This will be done later for R_COMMENT and R_RESTART records, as extra structures
+		 * are saved after the comment or the number of CPU.
+		 */
+		if ((record_hdr->record_type != R_COMMENT) && (record_hdr->record_type != R_RESTART) &&
+		    record_hdr->extra_next && (skip_extra_struct(ifd, endian_mismatch, arch_64) < 0))
+			return 2;
 	}
+	while ((record_hdr->record_type >= R_EXTRA_MIN) && (record_hdr->record_type <= R_EXTRA_MAX)) ;
 
 	return 0;
 }
@@ -1508,7 +1643,7 @@ __nr_t read_nr_value(int ifd, char *file, struct file_magic *file_magic,
 {
 	__nr_t value;
 
-	sa_fread(ifd, &value, sizeof(__nr_t), HARD_SIZE);
+	sa_fread(ifd, &value, sizeof(__nr_t), HARD_SIZE, UEOF_STOP);
 
 	/* Normalize endianness for file_activity structures */
 	if (endian_mismatch) {
@@ -1544,11 +1679,18 @@ __nr_t read_nr_value(int ifd, char *file, struct file_magic *file_magic,
  * @dfile	Name of system activity data file.
  * @file_magic	file_magic structure containing data read from file magic
  *		header.
+ * @oneof	Set to UEOF_CONT if an unexpected end of file should not make
+ *		sadf stop. Default behavior is to stop on unexpected EOF.
+ *
+ * RETURNS:
+ * 2 if an error has been encountered (e.g. unexpected EOF),
+ * 0 otherwise.
  ***************************************************************************
  */
-void read_file_stat_bunch(struct activity *act[], int curr, int ifd, int act_nr,
-			  struct file_activity *file_actlst, int endian_mismatch,
-			  int arch_64, char *dfile, struct file_magic *file_magic)
+int read_file_stat_bunch(struct activity *act[], int curr, int ifd, int act_nr,
+			 struct file_activity *file_actlst, int endian_mismatch,
+			 int arch_64, char *dfile, struct file_magic *file_magic,
+			 int oneof)
 {
 	int i, j, p;
 	struct file_activity *fal = file_actlst;
@@ -1584,6 +1726,8 @@ void read_file_stat_bunch(struct activity *act[], int curr, int ifd, int act_nr,
 				if (lseek(ifd, offset, SEEK_CUR) < offset) {
 					close(ifd);
 					perror("lseek");
+					if (oneof == UEOF_CONT)
+						return 2;
 					exit(2);
 				}
 			}
@@ -1620,8 +1764,10 @@ void read_file_stat_bunch(struct activity *act[], int curr, int ifd, int act_nr,
 		    (act[p]->msize > act[p]->fsize)) {
 
 			for (j = 0; j < (nr_value * act[p]->nr2); j++) {
-				sa_fread(ifd, (char *) act[p]->buf[curr] + j * act[p]->msize,
-					 (size_t) act[p]->fsize, HARD_SIZE);
+				if (sa_fread(ifd, (char *) act[p]->buf[curr] + j * act[p]->msize,
+					 (size_t) act[p]->fsize, HARD_SIZE, oneof) > 0)
+					/* Unexpected EOF */
+					return 2;
 			}
 		}
 		else if (nr_value > 0) {
@@ -1629,8 +1775,11 @@ void read_file_stat_bunch(struct activity *act[], int curr, int ifd, int act_nr,
 			 * Note: If msize was smaller than fsize,
 			 * then it has been set to fsize in check_file_actlst().
 			 */
-			sa_fread(ifd, act[p]->buf[curr],
-				 (size_t) act[p]->fsize * (size_t) nr_value * (size_t) act[p]->nr2, HARD_SIZE);
+			if (sa_fread(ifd, act[p]->buf[curr],
+				 (size_t) act[p]->fsize * (size_t) nr_value * (size_t) act[p]->nr2,
+				 HARD_SIZE, oneof) > 0)
+				/* Unexpected EOF */
+				return 2;
 		}
 		else {
 			/* nr_value == 0: Nothing to read */
@@ -1647,11 +1796,14 @@ void read_file_stat_bunch(struct activity *act[], int curr, int ifd, int act_nr,
 
 		/* Remap structure's fields to those known by current sysstat version */
 		for (j = 0; j < (nr_value * act[p]->nr2); j++) {
-			remap_struct(act[p]->gtypes_nr, act[p]->ftypes_nr,
-				     (char *) act[p]->buf[curr] + j * act[p]->msize,
-				     act[p]->fsize, act[p]->msize, act[p]->msize);
+			if (remap_struct(act[p]->gtypes_nr, act[p]->ftypes_nr,
+					 (char *) act[p]->buf[curr] + j * act[p]->msize,
+					 act[p]->fsize, act[p]->msize, act[p]->msize) < 0)
+				return 2;
 		}
 	}
+
+	return 0;
 }
 
 /*
@@ -1728,8 +1880,7 @@ int sa_open_read_magic(int *fd, char *dfile, struct file_magic *file_magic,
 	    ((file_magic->sysstat_version == 10) && (file_magic->sysstat_patchlevel >= 3))) {
 		/* header_size field exists only for sysstat versions 10.3.1 and later */
 		if ((file_magic->header_size <= MIN_FILE_HEADER_SIZE) ||
-		    (file_magic->header_size > MAX_FILE_HEADER_SIZE) ||
-		    ((file_magic->header_size < FILE_HEADER_SIZE) && !ignore)) {
+		    (file_magic->header_size > MAX_FILE_HEADER_SIZE)) {
 #ifdef DEBUG
 			fprintf(stderr, "%s: header_size=%u\n",
 				__FUNCTION__, file_magic->header_size);
@@ -1788,7 +1939,7 @@ int sa_open_read_magic(int *fd, char *dfile, struct file_magic *file_magic,
  * @arch_64	TRUE if file's data come from a 64 bit machine.
  ***************************************************************************
  */
-void check_file_actlst(int *ifd, char *dfile, struct activity *act[], unsigned int flags,
+void check_file_actlst(int *ifd, char *dfile, struct activity *act[], uint64_t flags,
 		       struct file_magic *file_magic, struct file_header *file_hdr,
 		       struct file_activity **file_actlst, unsigned int id_seq[],
 		       int ignore, int *endian_mismatch, int *arch_64)
@@ -1796,6 +1947,8 @@ void check_file_actlst(int *ifd, char *dfile, struct activity *act[], unsigned i
 	int i, j, k, p;
 	struct file_activity *fal;
 	void *buffer = NULL;
+	size_t bh_size = FILE_HEADER_SIZE;
+	size_t ba_size = FILE_ACTIVITY_SIZE;
 
 	/* Open sa data file and read its magic structure */
 	if (sa_open_read_magic(ifd, dfile, file_magic, ignore, endian_mismatch, TRUE) < 0)
@@ -1822,17 +1975,22 @@ void check_file_actlst(int *ifd, char *dfile, struct activity *act[], unsigned i
 	}
 
 	/* Allocate buffer for file_header structure */
-	SREALLOC(buffer, char, file_magic->header_size);
+	if (file_magic->header_size > FILE_HEADER_SIZE) {
+		bh_size = file_magic->header_size;
+	}
+	SREALLOC(buffer, char, bh_size);
 
 	/* Read sa data file standard header and allocate activity list */
-	sa_fread(*ifd, buffer, (size_t) file_magic->header_size, HARD_SIZE);
+	sa_fread(*ifd, buffer, (size_t) file_magic->header_size, HARD_SIZE, UEOF_STOP);
 	/*
 	 * Data file header size (file_magic->header_size) may be greater or
 	 * smaller than FILE_HEADER_SIZE. Remap the fields of the file header
-	 * then copy its contents to the expected  structure.
+	 * then copy its contents to the expected structure.
 	 */
-	remap_struct(hdr_types_nr, file_magic->hdr_types_nr, buffer,
-		     file_magic->header_size, FILE_HEADER_SIZE, file_magic->header_size);
+	if (remap_struct(hdr_types_nr, file_magic->hdr_types_nr, buffer,
+			 file_magic->header_size, FILE_HEADER_SIZE, bh_size) < 0)
+		goto format_error;
+
 	memcpy(file_hdr, buffer, FILE_HEADER_SIZE);
 	free(buffer);
 	buffer = NULL;
@@ -1865,7 +2023,11 @@ void check_file_actlst(int *ifd, char *dfile, struct activity *act[], unsigned i
 		goto format_error;
 	}
 
-	SREALLOC(buffer, char, file_hdr->act_size);
+	/* Allocate buffer for file_activity structures */
+	if (file_hdr->act_size > FILE_ACTIVITY_SIZE) {
+		ba_size = file_hdr->act_size;
+	}
+	SREALLOC(buffer, char, ba_size);
 	SREALLOC(*file_actlst, struct file_activity, FILE_ACTIVITY_SIZE * file_hdr->sa_act_nr);
 	fal = *file_actlst;
 
@@ -1874,14 +2036,15 @@ void check_file_actlst(int *ifd, char *dfile, struct activity *act[], unsigned i
 	for (i = 0; i < file_hdr->sa_act_nr; i++, fal++) {
 
 		/* Read current file_activity structure from file */
-		sa_fread(*ifd, buffer, (size_t) file_hdr->act_size, HARD_SIZE);
+		sa_fread(*ifd, buffer, (size_t) file_hdr->act_size, HARD_SIZE, UEOF_STOP);
 		/*
 		* Data file_activity size (file_hdr->act_size) may be greater or
 		* smaller than FILE_ACTIVITY_SIZE. Remap the fields of the file's structure
 		* then copy its contents to the expected structure.
 		*/
-		remap_struct(act_types_nr, file_hdr->act_types_nr, buffer,
-			     file_hdr->act_size, FILE_ACTIVITY_SIZE, file_hdr->act_size);
+		if (remap_struct(act_types_nr, file_hdr->act_types_nr, buffer,
+			     file_hdr->act_size, FILE_ACTIVITY_SIZE, ba_size) < 0)
+			goto format_error;
 		memcpy(fal, buffer, FILE_ACTIVITY_SIZE);
 
 		/* Normalize endianness for file_activity structures */
@@ -1990,6 +2153,7 @@ void check_file_actlst(int *ifd, char *dfile, struct activity *act[], unsigned i
 	}
 
 	free(buffer);
+	buffer = NULL;
 
 	/* Check that at least one activity selected by the user is available in file */
 	for (i = 0; i < NR_ACT; i++) {
@@ -2021,6 +2185,13 @@ void check_file_actlst(int *ifd, char *dfile, struct activity *act[], unsigned i
 		exit(1);
 	}
 
+	/*
+	 * Check if there are some extra structures.
+	 * We will just skip them as they are unknown for now.
+	 */
+	if (file_hdr->extra_next && (skip_extra_struct(*ifd, *endian_mismatch, *arch_64) < 0))
+		goto format_error;
+
 	return;
 
 format_error:
@@ -2048,7 +2219,7 @@ format_error:
  ***************************************************************************
  */
 int parse_sar_opt(char *argv[], int *opt, struct activity *act[],
-		  unsigned int *flags, int caller)
+		  uint64_t *flags, int caller)
 {
 	int i, p;
 
@@ -2062,21 +2233,19 @@ int parse_sar_opt(char *argv[], int *opt, struct activity *act[],
 
 		case 'A':
 			select_all_activities(act);
+			*flags |= S_F_OPTION_A;
 
 			/*
-			 * Force '-P ALL -I ALL -r ALL -u ALL -F'.
+			 * Force '-r ALL -u ALL -F'.
 			 * Setting -F is compulsory because corresponding activity
 			 * has AO_MULTIPLE_OUTPUTS flag set.
+			 * -P ALL / -I ALL will be set only corresponding option has
+			 * not been exlicitly entered on the command line.
 			 */
 			p = get_activity_position(act, A_MEMORY, EXIT_IF_NOT_FOUND);
 			act[p]->opt_flags |= AO_F_MEMORY + AO_F_SWAP + AO_F_MEM_ALL;
 
-			p = get_activity_position(act, A_IRQ, EXIT_IF_NOT_FOUND);
-			memset(act[p]->bitmap->b_array, ~0,
-			       BITMAP_SIZE(act[p]->bitmap->b_size));
 			p = get_activity_position(act, A_CPU, EXIT_IF_NOT_FOUND);
-			memset(act[p]->bitmap->b_array, ~0,
-			       BITMAP_SIZE(act[p]->bitmap->b_size));
 			act[p]->opt_flags = AO_F_CPU_ALL;
 
 			p = get_activity_position(act, A_FS, EXIT_IF_NOT_FOUND);
@@ -2129,6 +2298,11 @@ int parse_sar_opt(char *argv[], int *opt, struct activity *act[],
 				return 1;
 			}
 			(*opt)++;
+			if (!strcmp(argv[*opt], K_SID)) {
+				*flags |= S_F_DEV_SID + S_F_DEV_PRETTY;
+				return 0;
+			}
+
 			if (strnlen(argv[*opt], MAX_FILE_LEN) >= MAX_FILE_LEN - 1)
 				return 1;
 
@@ -2403,13 +2577,14 @@ int parse_sar_n_opt(char *argv[], int *opt, struct activity *act[])
  * @act		Array of activities.
  *
  * OUT:
+ * @flags	Common flags and system state.
  * @act		Array of activities, with interrupts activity selected.
  *
  * RETURNS:
  * 0 on success, 1 otherwise.
  ***************************************************************************
  */
-int parse_sar_I_opt(char *argv[], int *opt, struct activity *act[])
+int parse_sar_I_opt(char *argv[], int *opt, uint64_t *flags, struct activity *act[])
 {
 	int p;
 
@@ -2422,6 +2597,7 @@ int parse_sar_I_opt(char *argv[], int *opt, struct activity *act[])
 			     act[p]->bitmap->b_size, K_SUM))
 			return 1;
 		(*opt)++;
+		*flags |= S_F_OPTION_I;
 		return 0;
 	}
 
@@ -2445,7 +2621,7 @@ int parse_sar_I_opt(char *argv[], int *opt, struct activity *act[])
  * 0 on success, 1 otherwise.
  ***************************************************************************
  */
-int parse_sa_P_opt(char *argv[], int *opt, unsigned int *flags, struct activity *act[])
+int parse_sa_P_opt(char *argv[], int *opt, uint64_t *flags, struct activity *act[])
 {
 	int p;
 
@@ -2456,12 +2632,44 @@ int parse_sa_P_opt(char *argv[], int *opt, unsigned int *flags, struct activity
 			     act[p]->bitmap->b_size, K_LOWERALL))
 			return 1;
 		(*opt)++;
+		*flags |= S_F_OPTION_P;
 		return 0;
 	}
 
 	return 1;
 }
 
+/*
+ ***************************************************************************
+ * If option -A has been used, force -P ALL -I ALL only if corresponding
+ * option has not been explicitly entered on the command line.
+ *
+ * IN:
+ * @flags	Common flags and system state.
+ *
+ * OUT:
+ * @act		Array of selected activities.
+ ***************************************************************************
+ */
+void set_bitmaps(struct activity *act[], uint64_t *flags)
+{
+	int p;
+
+	if (!USE_OPTION_P(*flags)) {
+		/* Force -P ALL */
+		p = get_activity_position(act, A_CPU, EXIT_IF_NOT_FOUND);
+		memset(act[p]->bitmap->b_array, ~0,
+		       BITMAP_SIZE(act[p]->bitmap->b_size));
+	}
+
+	if (!USE_OPTION_I(*flags)) {
+		/* Force -I ALL */
+		p = get_activity_position(act, A_IRQ, EXIT_IF_NOT_FOUND);
+		memset(act[p]->bitmap->b_array, ~0,
+		       BITMAP_SIZE(act[p]->bitmap->b_size));
+	}
+}
+
 /*
  ***************************************************************************
  * Count number of comma-separated values in arguments list. For example,
@@ -2643,7 +2851,7 @@ void replace_nonprintable_char(int ifd, char *comment)
 	int i;
 
 	/* Read comment */
-	sa_fread(ifd, comment, MAX_COMMENT_LEN, HARD_SIZE);
+	sa_fread(ifd, comment, MAX_COMMENT_LEN, HARD_SIZE, UEOF_STOP);
 	comment[MAX_COMMENT_LEN - 1] = '\0';
 
 	/* Replace non printable chars */
@@ -2684,41 +2892,27 @@ void replace_nonprintable_char(int ifd, char *comment)
  * 1 if an error was detected, or 0 otherwise.
  ***************************************************************************
 */
-int sa_get_record_timestamp_struct(unsigned int l_flags, struct record_header *record_hdr,
-				   struct tm *rectime, struct tm *loctime)
+int sa_get_record_timestamp_struct(uint64_t l_flags, struct record_header *record_hdr,
+				   struct tm *rectime)
 {
-	struct tm *ltm = NULL;
+	struct tm *ltm;
 	int rc = 0;
 
-	/* Fill localtime structure if given */
-	if (loctime) {
-		if ((ltm = localtime((const time_t *) &(record_hdr->ust_time))) != NULL) {
-			*loctime = *ltm;
-		}
-		else {
-			rc = 1;
-		}
-	}
-
-	/* Fill generic rectime structure */
-	if (PRINT_LOCAL_TIME(l_flags) && !ltm) {
-		/* Get local time if not already done */
-		ltm = localtime((const time_t *) &(record_hdr->ust_time));
-	}
+	/*
+	 * Fill generic rectime structure in local time.
+	 * Done so that we have some default values.
+	 */
+	ltm = localtime_r((const time_t *) &(record_hdr->ust_time), rectime);
 
 	if (!PRINT_LOCAL_TIME(l_flags) && !PRINT_TRUE_TIME(l_flags)) {
 		/*
 		 * Get time in UTC
 		 * (the user doesn't want local time nor time of file's creator).
 		 */
-		ltm = gmtime((const time_t *) &(record_hdr->ust_time));
+		ltm = gmtime_r((const time_t *) &(record_hdr->ust_time), rectime);
 	}
 
-	if (ltm) {
-		/* Done even in true time mode so that we have some default values */
-		*rectime = *ltm;
-	}
-	else {
+	if (!ltm) {
 		rc = 1;
 	}
 
@@ -2758,7 +2952,7 @@ int sa_get_record_timestamp_struct(unsigned int l_flags, struct record_header *r
  * 		been used.
  ***************************************************************************
 */
-void set_record_timestamp_string(unsigned int l_flags, struct record_header *record_hdr,
+void set_record_timestamp_string(uint64_t l_flags, struct record_header *record_hdr,
 				 char *cur_date, char *cur_time, int len, struct tm *rectime)
 {
 	/* Set cur_time date value */
@@ -2798,8 +2992,6 @@ void set_record_timestamp_string(unsigned int l_flags, struct record_header *rec
  * @rectime	Structure where timestamp (expressed in local time or in UTC
  *		depending on whether options -T/-t have been used or not) can
  *		be saved for current record.
- * @loctime	Structure where timestamp (expressed in local time) can be
- *		saved for current record. May be NULL.
  * @file	Name of file being read.
  * @tab		Number of tabulations to print.
  * @file_magic	file_magic structure filled with file magic header data.
@@ -2813,16 +3005,14 @@ void set_record_timestamp_string(unsigned int l_flags, struct record_header *rec
  * OUT:
  * @rectime	Structure where timestamp (expressed in local time or in UTC)
  *		has been saved.
- * @loctime	Structure where timestamp (expressed in local time) has been
- *		saved (if requested).
  *
  * RETURNS:
  * 1 if the record has been successfully displayed, and 0 otherwise.
  ***************************************************************************
  */
-int print_special_record(struct record_header *record_hdr, unsigned int l_flags,
+int print_special_record(struct record_header *record_hdr, uint64_t l_flags,
 			 struct tstamp *tm_start, struct tstamp *tm_end, int rtype, int ifd,
-			 struct tm *rectime, struct tm *loctime, char *file, int tab,
+			 struct tm *rectime, char *file, int tab,
 			 struct file_magic *file_magic, struct file_header *file_hdr,
 			 struct activity *act[], struct report_format *ofmt,
 			 int endian_mismatch, int arch_64)
@@ -2832,17 +3022,12 @@ int print_special_record(struct record_header *record_hdr, unsigned int l_flags,
 	int p;
 
 	/* Fill timestamp structure (rectime) for current record */
-	if (sa_get_record_timestamp_struct(l_flags, record_hdr, rectime, loctime))
+	if (sa_get_record_timestamp_struct(l_flags, record_hdr, rectime))
 		return 0;
 
-	/* If loctime is NULL, then use rectime for comparison */
-	if (!loctime) {
-		loctime = rectime;
-	}
-
 	/* The record must be in the interval specified by -s/-e options */
-	if ((tm_start->use && (datecmp(loctime, tm_start) < 0)) ||
-	    (tm_end->use && (datecmp(loctime, tm_end) > 0))) {
+	if ((tm_start->use && (datecmp(rectime, tm_start, FALSE) < 0)) ||
+	    (tm_end->use && (datecmp(rectime, tm_end, FALSE) > 0))) {
 		/* Will not display the special record */
 		dp = 0;
 	}
@@ -2874,13 +3059,17 @@ int print_special_record(struct record_header *record_hdr, unsigned int l_flags,
 			}
 		}
 
+		/* Ignore unknown extra structures if present */
+		if (record_hdr->extra_next && (skip_extra_struct(ifd, endian_mismatch, arch_64) < 0))
+			return 0;
+
 		if (!dp)
 			return 0;
 
 		if (*ofmt->f_restart) {
 			(*ofmt->f_restart)(&tab, F_MAIN, cur_date, cur_time,
 					   !PRINT_LOCAL_TIME(l_flags) &&
-					   !PRINT_TRUE_TIME(l_flags), file_hdr);
+					   !PRINT_TRUE_TIME(l_flags), file_hdr, record_hdr);
 		}
 	}
 	else if (rtype == R_COMMENT) {
@@ -2889,6 +3078,10 @@ int print_special_record(struct record_header *record_hdr, unsigned int l_flags,
 		/* Read and replace non printable chars in comment */
 		replace_nonprintable_char(ifd, file_comment);
 
+		/* Ignore unknown extra structures if present */
+		if (record_hdr->extra_next && (skip_extra_struct(ifd, endian_mismatch, arch_64) < 0))
+			return 0;
+
 		if (!dp || !DISPLAY_COMMENT(l_flags))
 			return 0;
 
@@ -2896,7 +3089,7 @@ int print_special_record(struct record_header *record_hdr, unsigned int l_flags,
 			(*ofmt->f_comment)(&tab, F_MAIN, cur_date, cur_time,
 					   !PRINT_LOCAL_TIME(l_flags) &&
 					   !PRINT_TRUE_TIME(l_flags), file_comment,
-					   file_hdr);
+					   file_hdr, record_hdr);
 		}
 	}
 
@@ -2928,7 +3121,7 @@ int print_special_record(struct record_header *record_hdr, unsigned int l_flags,
  ***************************************************************************
  */
 unsigned long long get_global_cpu_statistics(struct activity *a, int prev, int curr,
-					     unsigned int flags, unsigned char offline_cpu_bitmap[])
+					     uint64_t flags, unsigned char offline_cpu_bitmap[])
 {
 	int i;
 	unsigned long long tot_jiffies_c, tot_jiffies_p;
@@ -3075,7 +3268,7 @@ unsigned long long get_global_cpu_statistics(struct activity *a, int prev, int c
  ***************************************************************************
  */
 void get_global_soft_statistics(struct activity *a, int prev, int curr,
-				unsigned int flags, unsigned char offline_cpu_bitmap[])
+				uint64_t flags, unsigned char offline_cpu_bitmap[])
 {
 	int i;
 	struct stats_softnet *ssnc, *ssnp;
@@ -3136,15 +3329,20 @@ void get_global_soft_statistics(struct activity *a, int prev, int curr,
  * IN:
  * @major	Major number of the device.
  * @minor	Minor number of the device.
+ * @wwn		WWN identifier of the device (0 if unknown).
+ * @part_nr	Partition number (0 if unknown).
  * @flags	Flags for common options and system state.
  *
  * RETURNS:
  * The name of the device.
  ***************************************************************************
  */
-char *get_sa_devname(unsigned int major, unsigned int minor, unsigned int flags)
+char *get_sa_devname(unsigned int major, unsigned int minor, unsigned long long wwn[],
+		     unsigned int part_nr, uint64_t flags)
 {
 	char *dev_name = NULL, *persist_dev_name = NULL;
+	static char sid[64];
+	char xsid[32] = "", pn[16] = "";
 
 	if (DISPLAY_PERSIST_NAME_S(flags)) {
 		persist_dev_name = get_persistent_name_from_pretty(get_devname(major, minor, TRUE));
@@ -3154,7 +3352,17 @@ char *get_sa_devname(unsigned int major, unsigned int minor, unsigned int flags)
 		dev_name = persist_dev_name;
 	}
 	else {
-		if ((USE_PRETTY_OPTION(flags)) && (major == dm_major)) {
+		if ((USE_STABLE_ID(flags)) && (wwn[0] != 0)) {
+			if (wwn[1] != 0) {
+				sprintf(xsid, "%016llx", wwn[1]);
+			}
+			if (part_nr) {
+				sprintf(pn, "-%d", part_nr);
+			}
+			snprintf(sid, sizeof(sid), "%#016llx%s%s", wwn[0], xsid, pn);
+			dev_name = sid;
+		}
+		else if ((USE_PRETTY_OPTION(flags)) && (major == dm_major)) {
 			dev_name = transform_devmapname(major, minor);
 		}
 
